package com.openexchange.guard.drive;

import java.io.InputStream;
import java.util.ArrayList;
import org.apache.http.entity.ContentType;
import com.google.gson.JsonObject;
import com.openexchange.guard.database.RecipKey;
import com.openexchange.guard.encr.GuardKeys;
import com.openexchange.guard.ox.Api;
import com.openexchange.guard.ox.streaming.JsonStreamHandler;
import com.openexchange.guard.ox.streaming.PGPStreamEncryptionHandler;
import com.openexchange.guard.ox.streaming.StreamHandler;


/**
 * Class for encrypting a drive file
 */
public class FileEncrypter {

    private final Api api;

    /**
     * Constructor
     * @param api HTTP API access to drive
     */
    public FileEncrypter(Api api){
        this.api = api;
    }

    /**
     * Encrypts and signs an input stream and uploads the encrypted content to a new drive file
     *
     * @param inputStream The stream to the item to encrypt
     * @param fileName The name of the file to encrypt
     * @param fileType The content type of the file to encrypt
     * @param jsonStream The stream to a JSON object which contains meta data which should be written to the new file
     * @param jsonHandler A handler which can change the JSON data during processing the jsonStream
     * @param recipients A list of recipients to encrypt the file for
     * @param signingKey The key to sign the item with, or null
     *        if the item should not be signed
     * @param signingKeyPassword The password for using the signing key,
     *        or null if signingKey is null
     * @return The JSON response from the backend if the upload of the encrypted item was successful
     * @throws Exception On an error during encryption or down/uploading
     */
    public JsonObject Encrypt(InputStream inputStream,
                              String fileName,
                              String fileType,
                              InputStream jsonStream,
                              JsonStreamHandler jsonHandler,
                              ArrayList<RecipKey> recipients,
                              GuardKeys signingKey,
                              String signingKeyPassword,
                              boolean update) throws Exception {

        //Setup up the encryption handling using PGP as default encryption mode
        StreamHandler streamEncryptor = null;
        if(signingKey == null) {
            //encryption only
            streamEncryptor = new PGPStreamEncryptionHandler(recipients,false /* armored */);
            //guard encryption for testing purpose
            //You can enable this for testing purpose using the legacy guard encryption
            //streamEncryptor = new GuardStreamEcryptionHandler(recipients.get(0).userid,
            //                                                  recipients.get(0).cid);
        } else {
            //encryption and signing
            streamEncryptor =
                new PGPStreamEncryptionHandler(recipients,
                                               signingKey,
                                               signingKeyPassword,
                                               true /* armored */);
        }

        String newItemName = streamEncryptor.getNewName(fileName);
        ContentType newContentType =
            ContentType.create(streamEncryptor.getNewContentType(fileType, fileName));

        //Upload the stream while encrypting the data
        return api.UploadDriveStreamItem(inputStream,
                                         streamEncryptor,
                                         newItemName,
                                         newContentType,
                                         jsonStream,
                                         jsonHandler,
                                         update);
    }

    /**
     * Encrypts an input stream and uploads the encrypted content to a new drive file
     *
     * @param inputStream The stream to encrypt
     * @param fileName The name of the file to encrypt
     * @param fileType The content type of the file to encrypt
     * @param jsonStream The stream to a JSON object which contains meta data which should be written to the new file
     * @param jsonHandler A handler which can change the JSON data during processing the jsonStream
     * @param recipients A list of recipients to encrypt the file for
     * @return The JSON response from the backend if the upload of the encrypted item was successful
     * @throws Exception On an error during encryption or down/uploading
     */
    public JsonObject Encrypt(InputStream inputStream,
                              String fileName,
                              String fileType,
                              InputStream jsonStream,
                              JsonStreamHandler jsonHandler,
                              ArrayList<RecipKey> recipients,
                              boolean update) throws Exception {
        return Encrypt(inputStream,
                       fileName,
                       fileType,
                       jsonStream,
                       jsonHandler,
                       recipients,
                       null,  /*no signing key: do not sign */
                       null, /*no signing key password: do not sign */
                       update);
    }
}
